"""RapidFit - Build multi-task classifiers and augment classification datasets."""

__version__ = "0.1.4"

from rapidfit.augmenters import BaseAugmenter, LLMAugmenter
from rapidfit.classifiers import (
    BaseClassifier,
    ClassifierType,
    EncoderConfig,
    EvalConfig,
    HeadConfig,
    LossConfig,
    MultiheadClassifier,
    MultiheadConfig,
    TrainingConfig,
    create_classifier,
    export_to_onnx,
    quantize_onnx,
)
from rapidfit.io import DataSaver
from rapidfit.types import (
    AnalysisResult,
    AugmentResult,
    ClassifierConfig,
    ClassMetrics,
    ErrorSample,
    Prediction,
    Sample,
    SaveFormat,
    SeedData,
    TaskAnalysis,
    TaskResult,
    TaskStats,
    WriteMode,
)

__all__ = [
    "AnalysisResult",
    "AugmentResult",
    "BaseAugmenter",
    "BaseClassifier",
    "ClassifierConfig",
    "ClassifierType",
    "ClassMetrics",
    "DataSaver",
    "EncoderConfig",
    "ErrorSample",
    "EvalConfig",
    "HeadConfig",
    "LLMAugmenter",
    "LossConfig",
    "MultiheadClassifier",
    "MultiheadConfig",
    "Prediction",
    "Sample",
    "SaveFormat",
    "SeedData",
    "TaskAnalysis",
    "TaskResult",
    "TaskStats",
    "TrainingConfig",
    "WriteMode",
    "create_classifier",
    "export_to_onnx",
    "quantize_onnx",
]



"""Reusable neural network components."""

from rapidfit.classifiers.components.analysis import compute_task_analysis
from rapidfit.classifiers.components.heads import TaskHeads, build_head
from rapidfit.classifiers.components.losses import FocalLoss, TaskLoss
from rapidfit.classifiers.components.pooling import Pooler

__all__ = [
    "FocalLoss",
    "Pooler",
    "TaskHeads",
    "TaskLoss",
    "build_head",
    "compute_task_analysis",
]



"""Classifier module exports and factory."""

from enum import Enum

from rapidfit.classifiers.base import BaseClassifier
from rapidfit.classifiers.config import (
    DEFAULT_CONFIG,
    EncoderConfig,
    EvalConfig,
    HeadConfig,
    LossConfig,
    MultiheadConfig,
    TrainingConfig,
)
from rapidfit.classifiers.export import export_to_onnx, quantize_onnx
from rapidfit.classifiers.multihead import MultiheadClassifier
from rapidfit.types import ClassifierConfig


class ClassifierType(Enum):
    """Available classifier types."""
    MULTIHEAD = "multihead"


def create_classifier(
    classifier_type: ClassifierType | str = ClassifierType.MULTIHEAD,
    config: ClassifierConfig | MultiheadConfig | None = None,
) -> BaseClassifier:
    """
    Factory function to create classifiers.

    Args:
        classifier_type: Type of classifier to create.
        config: Optional configuration dict or MultiheadConfig.

    Returns:
        Configured classifier instance.
    """
    if isinstance(classifier_type, str):
        classifier_type = ClassifierType(classifier_type)

    classifiers = {
        ClassifierType.MULTIHEAD: MultiheadClassifier,
    }

    cls = classifiers.get(classifier_type)
    if not cls:
        raise ValueError(f"Unknown classifier type: {classifier_type}")

    return cls(config)


__all__ = [
    "BaseClassifier",
    "ClassifierType",
    "DEFAULT_CONFIG",
    "EncoderConfig",
    "EvalConfig",
    "HeadConfig",
    "LossConfig",
    "MultiheadClassifier",
    "MultiheadConfig",
    "TrainingConfig",
    "create_classifier",
    "export_to_onnx",
    "quantize_onnx",
]



"""Base augmenter interface."""

from abc import ABC, abstractmethod

from rapidfit.types import AugmentResult, SeedData


class BaseAugmenter(ABC):
    """Abstract base class for data augmenters."""

    @abstractmethod
    def augment(self, seed_data: SeedData) -> AugmentResult:
        """
        Augment the seed dataset.

        Args:
            seed_data: Mapping of task names to sample lists.

        Returns:
            Mapping of task names to results with file paths and stats.
        """
        pass



"""Error analysis utilities for classification tasks."""

from collections import defaultdict

import numpy as np

from rapidfit.types import ClassMetrics, ErrorSample, TaskAnalysis


def compute_task_analysis(
    texts: list[str],
    y_true: list[int],
    y_pred: list[int],
    confidences: list[float],
    labels: list[str],
) -> TaskAnalysis:
    """Compute complete analysis for a single task."""
    n_labels = len(labels)
    y_true_arr = np.array(y_true)
    y_pred_arr = np.array(y_pred)

    # Confusion matrix
    matrix = [[0] * n_labels for _ in range(n_labels)]
    for t, p in zip(y_true, y_pred):
        matrix[t][p] += 1

    # Per-class metrics
    class_metrics = {}
    for i, label in enumerate(labels):
        tp = matrix[i][i]
        fp = sum(matrix[j][i] for j in range(n_labels) if j != i)
        fn = sum(matrix[i][j] for j in range(n_labels) if j != i)
        support = sum(matrix[i])

        precision = tp / (tp + fp) if (tp + fp) > 0 else 0.0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0.0
        f1 = 2 * precision * recall / (precision + recall) if (precision + recall) > 0 else 0.0

        class_metrics[label] = ClassMetrics(
            precision=round(precision, 4),
            recall=round(recall, 4),
            f1=round(f1, 4),
            support=support,
        )

    # Collect errors
    errors = []
    for i, (t, p) in enumerate(zip(y_true, y_pred)):
        if t != p:
            errors.append(ErrorSample(
                text=texts[i],
                true_label=labels[t],
                predicted_label=labels[p],
                confidence=round(confidences[i], 4),
            ))

    # Sort by confidence descending (high-confidence errors first)
    errors.sort(key=lambda e: e["confidence"], reverse=True)

    return TaskAnalysis(
        accuracy=round(float((y_true_arr == y_pred_arr).mean()), 4),
        class_metrics=class_metrics,
        confusion_matrix=matrix,
        labels=labels,
        errors=errors,
    )



"""ONNX export and quantization utilities."""

import json
from pathlib import Path

import torch
import torch.nn as nn


class _ONNXInferenceModel(nn.Module):
    """Lightweight inference wrapper for ONNX export."""

    def __init__(self, encoder, pooler, dropout, head) -> None:
        super().__init__()
        self.encoder = encoder
        self.pooler = pooler
        self.dropout = dropout
        self.head = head

    def forward(self, input_ids: torch.Tensor, attention_mask: torch.Tensor):
        outputs = self.encoder(input_ids=input_ids, attention_mask=attention_mask)
        pooled = self.pooler(outputs.last_hidden_state, attention_mask)
        pooled = self.dropout(pooled)
        return self.head(pooled)


def export_to_onnx(
    model: nn.Module,
    task: str,
    output_path: str | Path,
    max_length: int = 128,
    opset_version: int = 17,
    external_data: bool | None = None,
) -> Path:
    """
    Export a single task to ONNX format.

    Args:
        model: The _MultiTaskModel instance.
        task: Task name to export.
        output_path: Directory to save the ONNX file.
        max_length: Maximum sequence length for dummy input.
        opset_version: ONNX opset version.
        external_data: Store weights externally for large models (>2GB).
                       None=auto-detect, True=force, False=disable.

    Returns:
        Path to the exported ONNX file.
    """
    try:
        import onnx
    except ImportError:
        raise ImportError("Install export dependencies: pip install rapidfit[export]")

    if task not in model.task_heads:
        raise ValueError(f"Unknown task: {task}")

    output_path = Path(output_path)
    output_path.mkdir(parents=True, exist_ok=True)

    inference_model = _ONNXInferenceModel(
        encoder=model.encoder,
        pooler=model.pooler,
        dropout=nn.Dropout(0.0),  # No dropout at inference
        head=model.task_heads[task],
    )
    inference_model.eval()

    device = next(model.parameters()).device
    dummy_input_ids = torch.ones(1, max_length, dtype=torch.long, device=device)
    dummy_attention_mask = torch.ones(1, max_length, dtype=torch.long, device=device)

    onnx_path = output_path / f"{task}.onnx"

    if external_data is None:
        param_bytes = sum(p.numel() * p.element_size() for p in model.parameters())
        external_data = param_bytes > 2 * 1024**3

    with torch.no_grad():
        torch.onnx.export(
            inference_model,
            (dummy_input_ids, dummy_attention_mask),
            str(onnx_path),
            input_names=["input_ids", "attention_mask"],
            output_names=["logits"],
            dynamic_axes={
                "input_ids": {0: "batch", 1: "seq"},
                "attention_mask": {0: "batch", 1: "seq"},
                "logits": {0: "batch"},
            },
            opset_version=opset_version,
            do_constant_folding=True,
        )

    if external_data:
        onnx_model = onnx.load(str(onnx_path), load_external_data=False)
        onnx.save(
            onnx_model,
            str(onnx_path),
            save_as_external_data=True,
            location=f"{task}_weights.bin",
            all_tensors_to_one_file=True,
        )
        onnx.checker.check_model(str(onnx_path))
    else:
        onnx_model = onnx.load(str(onnx_path))
        onnx.checker.check_model(onnx_model)

    config_path = output_path / f"{task}_config.json"
    config_path.write_text(json.dumps({
        "id2label": {str(k): v for k, v in model.task_id2label[task].items()},
        "label2id": model.task_label2id[task],
    }, indent=2))

    return onnx_path


def quantize_onnx(
    onnx_path: str | Path,
    output_path: str | Path | None = None,
) -> Path:
    """
    Apply INT8 dynamic quantization to an ONNX model.

    Args:
        onnx_path: Path to the ONNX model.
        output_path: Output path. Defaults to <name>_int8.onnx.

    Returns:
        Path to the quantized model.
    """
    try:
        import onnx
        from onnxruntime.quantization import quantize_dynamic, QuantType
    except ImportError:
        raise ImportError("Install export dependencies: pip install rapidfit[export]")

    onnx_path = Path(onnx_path)
    if output_path is None:
        output_path = onnx_path.with_stem(f"{onnx_path.stem}_int8")
    else:
        output_path = Path(output_path)

    # Use standard ONNX shape inference (handles dynamic axes better)
    preprocessed = onnx_path.with_stem(f"{onnx_path.stem}_prep")
    model = onnx.load(str(onnx_path))
    model = onnx.shape_inference.infer_shapes(model)
    onnx.save(model, str(preprocessed))

    quantize_dynamic(
        model_input=str(preprocessed),
        model_output=str(output_path),
        weight_type=QuantType.QInt8,
    )

    preprocessed.unlink()
    return output_path



"""Multihead classifier implementation."""

import os
import random
from collections import Counter, defaultdict
from dataclasses import asdict
from pathlib import Path
from typing import Any

import numpy as np
import torch
import torch.nn as nn
from datasets import Dataset, DatasetDict
from rich.console import Console
from rich.table import Table
from sklearn.model_selection import train_test_split
from transformers import (
    AutoModel,
    AutoTokenizer,
    EarlyStoppingCallback,
    Trainer,
    TrainingArguments,
)

from rapidfit.classifiers.base import BaseClassifier
from rapidfit.classifiers.components import Pooler, TaskHeads, TaskLoss, compute_task_analysis
from rapidfit.classifiers.config import (
    DEFAULT_CONFIG,
    HeadConfig,
    LossConfig,
    MultiheadConfig,
)
from rapidfit.types import AnalysisResult, AugmentResult, Prediction, SeedData

console = Console()


class _MultiTaskModel(nn.Module):
    """Neural network with shared encoder and task-specific heads."""

    def __init__(
        self,
        model_name: str,
        task_num_labels: dict[str, int],
        task_label2id: dict[str, dict[str, int]],
        task_id2label: dict[str, dict[int, str]],
        head_config: HeadConfig,
        loss_config: LossConfig,
        pooling: str = "mean",
    ) -> None:
        super().__init__()
        self.encoder = AutoModel.from_pretrained(model_name)
        self.task_num_labels = task_num_labels
        self.task_label2id = task_label2id
        self.task_id2label = task_id2label

        hidden_size = self.encoder.config.hidden_size
        self.pooler = Pooler(pooling)
        self.dropout = nn.Dropout(head_config.dropout)
        self.task_heads = TaskHeads(hidden_size, task_num_labels, head_config)
        self.task_loss = TaskLoss(loss_config)

        self._head_config = head_config
        self._loss_config = loss_config
        self._pooling = pooling

    def set_class_weights(self, weights: dict[str, torch.Tensor]) -> None:
        self.task_loss.set_class_weights(weights)

    def freeze_encoder(self) -> None:
        for param in self.encoder.parameters():
            param.requires_grad = False

    def unfreeze_encoder(self, num_layers: int | None = None) -> None:
        if num_layers is None:
            for param in self.encoder.parameters():
                param.requires_grad = True
            return

        for param in self.encoder.parameters():
            param.requires_grad = False

        if hasattr(self.encoder, "encoder") and hasattr(self.encoder.encoder, "layer"):
            layers = self.encoder.encoder.layer
            for i in range(max(0, len(layers) - num_layers), len(layers)):
                for param in layers[i].parameters():
                    param.requires_grad = True
        else:
            for param in self.encoder.parameters():
                param.requires_grad = True

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        task_name: str | None = None,
        labels: torch.Tensor | None = None,
    ) -> dict[str, Any]:
        outputs = self.encoder(input_ids=input_ids, attention_mask=attention_mask)
        pooled = self.pooler(outputs.last_hidden_state, attention_mask)
        pooled = self.dropout(pooled)
        logits = self.task_heads[task_name](pooled)

        loss = None
        if labels is not None:
            loss = self.task_loss.compute(logits, labels, task_name)

        return {"loss": loss, "logits": logits}

    def save_pretrained(self, path: str) -> None:
        os.makedirs(path, exist_ok=True)
        self.encoder.save_pretrained(os.path.join(path, "encoder"))

        torch.save(
            {
                "task_heads": self.task_heads.state_dict(),
                "task_num_labels": self.task_num_labels,
                "task_label2id": self.task_label2id,
                "task_id2label": self.task_id2label,
                "head_config": asdict(self._head_config),
                "loss_config": asdict(self._loss_config),
                "pooling": self._pooling,
                "class_weights": self.task_loss.state_dict(),
            },
            os.path.join(path, "task_heads.pt"),
        )

    @classmethod
    def from_pretrained(cls, path: str) -> "_MultiTaskModel":
        data = torch.load(os.path.join(path, "task_heads.pt"), map_location="cpu")
        model = cls(
            model_name=os.path.join(path, "encoder"),
            task_num_labels=data["task_num_labels"],
            task_label2id=data["task_label2id"],
            task_id2label=data["task_id2label"],
            head_config=HeadConfig(**data.get("head_config", {})),
            loss_config=LossConfig(**data.get("loss_config", {})),
            pooling=data.get("pooling", "mean"),
        )
        model.task_heads.load_state_dict(data["task_heads"])
        if "class_weights" in data:
            model.task_loss.load_state_dict(data["class_weights"])
        return model


class _TaskGroupedSampler:
    """Sampler ensuring each batch contains samples from a single task."""

    def __init__(
        self,
        dataset: Dataset,
        batch_size: int,
        shuffle: bool = True,
        sampling: str = "proportional",
    ) -> None:
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.sampling = sampling
        self.task_indices: dict[str, list[int]] = defaultdict(list)

        for idx in range(len(dataset)):
            self.task_indices[dataset[idx]["task"]].append(idx)

    def __iter__(self):
        batches = []
        task_samples = self._get_task_samples()

        for task, indices in task_samples.items():
            if self.shuffle:
                random.shuffle(indices)
            for i in range(0, len(indices), self.batch_size):
                batches.append(indices[i : i + self.batch_size])

        if self.shuffle:
            random.shuffle(batches)
        yield from batches

    def _get_task_samples(self) -> dict[str, list[int]]:
        if self.sampling == "proportional":
            return {t: list(idx) for t, idx in self.task_indices.items()}

        counts = {t: len(idx) for t, idx in self.task_indices.items()}
        if self.sampling == "equal":
            target = max(counts.values())
        else:  # sqrt
            target = int(max(counts.values()) ** 0.5 * min(counts.values()) ** 0.5)

        result = {}
        for task, indices in self.task_indices.items():
            if len(indices) >= target:
                result[task] = indices[:target]
            else:
                repeats = (target // len(indices)) + 1
                result[task] = (indices * repeats)[:target]
        return result

    def __len__(self) -> int:
        samples = self._get_task_samples()
        return sum(
            (len(idx) + self.batch_size - 1) // self.batch_size
            for idx in samples.values()
        )


class _MultiTaskCollator:
    """Data collator handling task field and label validation."""

    def __init__(self, tokenizer: Any, task_num_labels: dict[str, int]) -> None:
        self.tokenizer = tokenizer
        self.task_num_labels = task_num_labels

    def __call__(self, features: list[dict]) -> dict[str, Any]:
        tasks = [f.pop("task") for f in features]
        labels = [f.pop("label") for f in features]

        batch = self.tokenizer.pad(features, padding=True, return_tensors="pt")
        batch["task"] = tasks
        batch["labels"] = torch.tensor(labels, dtype=torch.long)
        return batch


class _MultiTaskTrainer(Trainer):
    """Trainer with task-grouped batching and configurable sampling."""

    def __init__(self, *args, task_sampling: str = "proportional", **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.task_sampling = task_sampling

    def compute_loss(self, model, inputs, return_outputs=False, **kwargs):
        labels = inputs.pop("labels")
        task_name = inputs.pop("task")[0]

        outputs = model(
            input_ids=inputs["input_ids"],
            attention_mask=inputs["attention_mask"],
            task_name=task_name,
            labels=labels,
        )
        return (outputs["loss"], outputs) if return_outputs else outputs["loss"]

    def _get_dataloader(self, dataset: Dataset, batch_size: int, shuffle: bool):
        from torch.utils.data import DataLoader

        sampler = _TaskGroupedSampler(
            dataset, batch_size, shuffle, self.task_sampling
        )
        return DataLoader(
            dataset,
            batch_sampler=sampler,
            collate_fn=self.data_collator,
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
        )

    def get_train_dataloader(self):
        return self._get_dataloader(
            self.train_dataset, self.args.per_device_train_batch_size, True
        )

    def get_eval_dataloader(self, eval_dataset=None):
        return self._get_dataloader(
            eval_dataset or self.eval_dataset,
            self.args.per_device_eval_batch_size,
            False,
        )


class MultiheadClassifier(BaseClassifier):
    """
    Classifier with shared encoder and task-specific heads.

    Each task gets its own classification head while sharing
    the underlying encoder for efficient multi-task learning.
    """

    def __init__(self, config: MultiheadConfig | dict | None = None) -> None:
        super().__init__(config)
        if config is None:
            self._cfg = DEFAULT_CONFIG
        elif isinstance(config, dict):
            self._cfg = MultiheadConfig.from_dict(config)
        else:
            self._cfg = config

        self._model: _MultiTaskModel | None = None
        self._tokenizer = None
        self._label_maps: dict[str, dict[str, int]] = {}
        self._id_maps: dict[str, dict[int, str]] = {}
        self._task_num_labels: dict[str, int] = {}
        self._splits: DatasetDict | None = None

    def train(
        self,
        data: SeedData | AugmentResult | None = None,
        data_save_dir: str | None = None,
    ) -> None:
        """Train multihead classifier on data."""
        samples = self._resolve_data(data, data_save_dir)
        if not samples:
            raise ValueError("No training data provided")

        self._build_label_maps(samples)
        self._log_data_stats(samples)

        datasets, task_num_labels = self._prepare_datasets(samples)
        self._task_num_labels = task_num_labels
        self._splits = datasets

        cfg = self._cfg
        self._tokenizer = AutoTokenizer.from_pretrained(cfg.training.model_name)
        self._model = _MultiTaskModel(
            model_name=cfg.training.model_name,
            task_num_labels=task_num_labels,
            task_label2id=self._label_maps,
            task_id2label=self._id_maps,
            head_config=cfg.head,
            loss_config=cfg.loss,
            pooling=cfg.pooling,
        )

        if cfg.loss.use_class_weights:
            weights = self._compute_class_weights(datasets["train"], task_num_labels)
            self._model.set_class_weights(weights)

        tokenized = datasets.map(
            lambda x: self._tokenizer(
                x["text"],
                padding="max_length",
                truncation=True,
                max_length=cfg.training.max_length,
            ),
            batched=True,
            remove_columns=["text"],
        )

        collator = _MultiTaskCollator(self._tokenizer, task_num_labels)

        if cfg.encoder.freeze_epochs > 0:
            self._train_phase(
                tokenized,
                collator,
                epochs=cfg.encoder.freeze_epochs,
                lr=cfg.training.learning_rate * 10,
                phase_name="heads-only",
                freeze_encoder=True,
            )

        self._train_phase(
            tokenized,
            collator,
            epochs=cfg.training.epochs,
            lr=cfg.training.learning_rate,
            phase_name="full",
            freeze_encoder=False,
        )

        self._is_trained = True
        console.print("[green]Training complete![/green]")

    def _train_phase(
        self,
        datasets: DatasetDict,
        collator: _MultiTaskCollator,
        epochs: int,
        lr: float,
        phase_name: str,
        freeze_encoder: bool,
    ) -> None:
        console.print(f"\n[bold cyan]Phase: {phase_name}[/bold cyan]")
        cfg = self._cfg

        if freeze_encoder:
            self._model.freeze_encoder()
        else:
            self._model.unfreeze_encoder(num_layers=cfg.encoder.unfreeze_layers)

        encoder_lr = lr * cfg.encoder.lr_multiplier if not freeze_encoder else lr

        args = TrainingArguments(
            output_dir=os.path.join(cfg.training.output_dir, phase_name),
            num_train_epochs=epochs,
            per_device_train_batch_size=cfg.training.batch_size,
            per_device_eval_batch_size=cfg.training.batch_size * 2,
            learning_rate=encoder_lr if freeze_encoder else lr,
            weight_decay=cfg.training.weight_decay,
            warmup_ratio=cfg.training.warmup_ratio,
            gradient_accumulation_steps=cfg.training.gradient_accumulation_steps,
            max_grad_norm=cfg.training.max_grad_norm,
            lr_scheduler_type="cosine",
            eval_strategy="epoch",
            save_strategy="epoch",
            load_best_model_at_end=True,
            metric_for_best_model=cfg.eval.metric,
            greater_is_better=(cfg.eval.metric != "loss"),
            save_total_limit=2,
            logging_steps=50,
            remove_unused_columns=False,
            fp16=torch.cuda.is_available(),
            report_to="none",
        )

        callbacks = []
        if cfg.eval.patience > 0:
            callbacks.append(
                EarlyStoppingCallback(
                    early_stopping_patience=cfg.eval.patience,
                    early_stopping_threshold=cfg.eval.min_delta,
                )
            )

        trainer = _MultiTaskTrainer(
            model=self._model,
            args=args,
            train_dataset=datasets["train"],
            eval_dataset=datasets["validation"],
            data_collator=collator,
            compute_metrics=self._compute_metrics,
            callbacks=callbacks,
            task_sampling=cfg.task_sampling,
        )

        trainer.train()

    @staticmethod
    def _compute_metrics(eval_pred) -> dict[str, float]:
        logits, labels = eval_pred
        preds = np.argmax(logits, axis=-1)
        return {"accuracy": (preds == labels).mean()}

    def predict(self, texts: list[str], task: str) -> list[Prediction]:
        """Predict labels for texts using task-specific head."""
        if not self._is_trained:
            raise RuntimeError("Model must be trained before prediction")
        if task not in self._model.task_heads:
            raise ValueError(f"Unknown task: {task}")

        self._model.eval()
        device = next(self._model.parameters()).device

        inputs = self._tokenizer(
            texts,
            padding=True,
            truncation=True,
            max_length=self._cfg.training.max_length,
            return_tensors="pt",
        ).to(device)

        with torch.no_grad():
            outputs = self._model(
                input_ids=inputs["input_ids"],
                attention_mask=inputs["attention_mask"],
                task_name=task,
            )

        probs = torch.softmax(outputs["logits"], dim=-1)
        confidences, indices = probs.max(dim=-1)
        id2label = self._id_maps[task]

        return [
            {"label": id2label[idx.item()], "confidence": conf.item()}
            for idx, conf in zip(indices, confidences)
        ]

    def predict_all_tasks(self, texts: list[str]) -> dict[str, list[Prediction]]:
        """Predict all tasks for given texts."""
        return {task: self.predict(texts, task) for task in self._model.task_heads}

    def analyze(
        self,
        split: str = "test",
        tasks: list[str] | None = None,
    ) -> AnalysisResult:
        """
        Analyze model performance on stored data splits.

        Args:
            split: Which split to analyze ('train', 'validation', 'test').
            tasks: Tasks to analyze. Defaults to all tasks.

        Returns:
            Analysis results with metrics, confusion matrix, and error samples.
        """
        if not self._is_trained:
            raise RuntimeError("Model must be trained before analysis")
        if self._splits is None:
            raise RuntimeError("No stored splits available. Train the model first.")
        if split not in self._splits:
            raise ValueError(f"Unknown split: {split}")

        dataset = self._splits[split]
        tasks = tasks or list(self._model.task_heads.keys())
        result: AnalysisResult = {"tasks": {}}

        for task in tasks:
            indices = [i for i, item in enumerate(dataset) if item["task"] == task]
            if not indices:
                continue

            texts = [dataset[i]["text"] for i in indices]
            y_true = [dataset[i]["label"] for i in indices]
            labels = list(self._id_maps[task].values())

            predictions = self.predict(texts, task)
            y_pred = [self._label_maps[task][p["label"]] for p in predictions]
            confidences = [p["confidence"] for p in predictions]

            result["tasks"][task] = compute_task_analysis(
                texts, y_true, y_pred, confidences, labels
            )

        return result

    def display(self, result: AnalysisResult, max_errors: int = 10) -> None:
        """
        Display analysis results in console.

        Args:
            result: Analysis result from analyze().
            max_errors: Maximum error samples to show per task.
        """
        for task, analysis in result["tasks"].items():
            console.print(f"\n[bold cyan]Task: {task}[/bold cyan]")
            console.print(f"Accuracy: {analysis['accuracy']:.2%}")

            # Class metrics table
            metrics_table = Table(title="Per-Class Metrics")
            metrics_table.add_column("Label", style="cyan")
            metrics_table.add_column("Precision", justify="right")
            metrics_table.add_column("Recall", justify="right")
            metrics_table.add_column("F1", justify="right")
            metrics_table.add_column("Support", justify="right")

            for label, m in analysis["class_metrics"].items():
                metrics_table.add_row(
                    label,
                    f"{m['precision']:.2%}",
                    f"{m['recall']:.2%}",
                    f"{m['f1']:.2%}",
                    str(m["support"]),
                )
            console.print(metrics_table)

            # Confusion matrix
            labels = analysis["labels"]
            matrix = analysis["confusion_matrix"]
            cm_table = Table(title="Confusion Matrix")
            cm_table.add_column("", style="bold")
            for lbl in labels:
                cm_table.add_column(lbl[:8], justify="right")

            for i, row_label in enumerate(labels):
                row = [row_label[:8]] + [str(matrix[i][j]) for j in range(len(labels))]
                cm_table.add_row(*row)
            console.print(cm_table)

            # Error samples
            errors = analysis["errors"][:max_errors]
            if errors:
                err_table = Table(title=f"Top {len(errors)} Errors (by confidence)")
                err_table.add_column("Text", max_width=50)
                err_table.add_column("True", style="green")
                err_table.add_column("Pred", style="red")
                err_table.add_column("Conf", justify="right")

                for e in errors:
                    text = e["text"][:47] + "..." if len(e["text"]) > 50 else e["text"]
                    err_table.add_row(
                        text, e["true_label"], e["predicted_label"], f"{e['confidence']:.2%}"
                    )
                console.print(err_table)

    def save(self, path: str | Path) -> None:
        """Save model to disk."""
        if not self._is_trained:
            raise RuntimeError("Cannot save untrained model")

        path = Path(path)
        self._model.save_pretrained(str(path))
        self._tokenizer.save_pretrained(str(path / "encoder"))
        console.print(f"[green]Model saved to {path}[/green]")

    def load(self, path: str | Path) -> None:
        """Load model from disk."""
        path = Path(path)
        self._model = _MultiTaskModel.from_pretrained(str(path))
        self._tokenizer = AutoTokenizer.from_pretrained(str(path / "encoder"))
        self._label_maps = self._model.task_label2id
        self._id_maps = self._model.task_id2label
        self._task_num_labels = self._model.task_num_labels
        self._is_trained = True
        console.print(f"[green]Model loaded from {path}[/green]")

    def _build_label_maps(self, data: SeedData) -> None:
        for task, samples in data.items():
            labels = sorted({s["label"] for s in samples})
            self._label_maps[task] = {lbl: i for i, lbl in enumerate(labels)}
            self._id_maps[task] = {i: lbl for i, lbl in enumerate(labels)}

    def _prepare_datasets(
        self, data: SeedData
    ) -> tuple[DatasetDict, dict[str, int]]:
        rows = []
        task_num_labels = {}

        for task, samples in data.items():
            label2id = self._label_maps[task]
            task_num_labels[task] = len(label2id)
            for s in samples:
                if s["label"] in label2id:
                    rows.append({
                        "text": s["text"],
                        "label": label2id[s["label"]],
                        "task": task,
                    })

        stratify_keys = [f"{r['task']}_{r['label']}" for r in rows]
        cfg = self._cfg

        try:
            train, test = train_test_split(
                rows,
                test_size=cfg.training.test_size,
                random_state=42,
                stratify=stratify_keys,
            )
            train_keys = [f"{r['task']}_{r['label']}" for r in train]
            train, val = train_test_split(
                train,
                test_size=cfg.training.val_size,
                random_state=42,
                stratify=train_keys,
            )
        except ValueError:
            task_keys = [r["task"] for r in rows]
            train, test = train_test_split(
                rows, test_size=cfg.training.test_size, random_state=42, stratify=task_keys
            )
            train_keys = [r["task"] for r in train]
            train, val = train_test_split(
                train, test_size=cfg.training.val_size, random_state=42, stratify=train_keys
            )

        console.print(
            f"[dim]Splits: train={len(train)}, val={len(val)}, test={len(test)}[/dim]"
        )

        return (
            DatasetDict({
                "train": Dataset.from_list(train),
                "validation": Dataset.from_list(val),
                "test": Dataset.from_list(test),
            }),
            task_num_labels,
        )

    def _compute_class_weights(
        self, dataset: Dataset, task_num_labels: dict[str, int]
    ) -> dict[str, torch.Tensor]:
        weights = {}
        for task, num_labels in task_num_labels.items():
            counts = Counter(
                item["label"] for item in dataset if item["task"] == task
            )
            if not counts:
                continue

            total = sum(counts.values())
            w = torch.zeros(num_labels)
            for label_id in range(num_labels):
                count = counts.get(label_id, 1)
                w[label_id] = total / (num_labels * count)

            w = w / w.mean()
            weights[task] = torch.clamp(w, min=0.5, max=3.0)

        return weights

    def _log_data_stats(self, data: SeedData) -> None:
        table = Table(title="Dataset Statistics")
        table.add_column("Task", style="cyan")
        table.add_column("Samples", justify="right")
        table.add_column("Labels", justify="right")

        for task, samples in data.items():
            labels = {s["label"] for s in samples}
            table.add_row(task, str(len(samples)), str(len(labels)))

        console.print(table)

    @property
    def tasks(self) -> list[str]:
        """Get list of available tasks."""
        return list(self._label_maps.keys())

    def get_labels(self, task: str) -> list[str]:
        """Get label names for a task."""
        if task not in self._label_maps:
            raise ValueError(f"Unknown task: {task}")
        return list(self._label_maps[task].keys())

    def export_onnx(
        self,
        path: str | Path,
        tasks: list[str] | None = None,
        quantize: bool = False,
        external_data: bool | None = None,
    ) -> dict[str, Path]:
        """
        Export model to ONNX format.

        Args:
            path: Output directory.
            tasks: Tasks to export. Defaults to all tasks.
            quantize: Apply INT8 quantization after export.
            external_data: Store weights externally for large models (>2GB).
                           None=auto-detect, True=force, False=disable.

        Returns:
            Dict mapping task names to ONNX file paths.
        """
        if not self._is_trained:
            raise RuntimeError("Model must be trained before export")

        from rapidfit.classifiers.export import export_to_onnx, quantize_onnx

        tasks = tasks or list(self._model.task_heads.keys())
        results = {}

        for task in tasks:
            onnx_path = export_to_onnx(
                self._model,
                task,
                path,
                max_length=self._cfg.training.max_length,
                external_data=external_data,
            )
            if quantize:
                onnx_path = quantize_onnx(onnx_path)
            results[task] = onnx_path
            console.print(f"[green]Exported {task} → {onnx_path}[/green]")

        return results



"""Type definitions for RapidFit."""

from enum import Enum
from typing import TypedDict


class SaveFormat(Enum):
    """Supported output formats for saving data."""
    JSON = "json"
    JSONL = "jsonl"
    CSV = "csv"


class WriteMode(Enum):
    """Mode for handling existing data files."""
    OVERWRITE = "overwrite"
    APPEND = "append"


class Sample(TypedDict):
    """A single classification sample."""
    text: str
    label: str


class LengthStats(TypedDict):
    """Length statistics for samples."""
    min: int
    max: int
    avg: int


class EdgeCases(TypedDict):
    """Edge case generation instructions."""
    short: str
    long: str
    noisy: str


class ClassInstruction(TypedDict):
    """Generation instructions for a class."""
    languages: dict[str, float]
    style: str
    length: LengthStats
    do: list[str]
    avoid: list[str]
    patterns: str
    edge_cases: EdgeCases


class TaskStats(TypedDict):
    """Statistics for a generated task."""
    total: int
    labels: dict[str, int]


class TaskResult(TypedDict):
    """Result for a single task augmentation."""
    path: str
    stats: TaskStats


class Prediction(TypedDict):
    """A single prediction result."""
    label: str
    confidence: float


class ErrorSample(TypedDict):
    """A misclassified sample with details."""
    text: str
    true_label: str
    predicted_label: str
    confidence: float


class ClassMetrics(TypedDict):
    """Per-class performance metrics."""
    precision: float
    recall: float
    f1: float
    support: int


class TaskAnalysis(TypedDict):
    """Analysis results for a single task."""
    accuracy: float
    class_metrics: dict[str, ClassMetrics]
    confusion_matrix: list[list[int]]
    labels: list[str]
    errors: list[ErrorSample]


class AnalysisResult(TypedDict):
    """Complete analysis across all tasks."""
    tasks: dict[str, TaskAnalysis]


SeedData = dict[str, list[Sample]]
"""Mapping of task names to lists of samples."""

AugmentResult = dict[str, TaskResult]
"""Mapping of task names to augmentation results."""

TaskPrompts = dict[str, dict[str, ClassInstruction]]
"""Mapping of task names to class instructions."""

ClassifierConfig = dict[str, any]
"""Configuration dictionary for classifiers."""



"""Data I/O utilities for saving and loading datasets."""

import csv
import json
from pathlib import Path

from rapidfit.types import Sample, SaveFormat


class DataSaver:
    """Handles saving and loading classification data in various formats."""

    def __init__(
        self,
        save_path: str = "./saved",
        save_format: SaveFormat | str = SaveFormat.JSON,
    ) -> None:
        self._path = Path(save_path)
        self._format = SaveFormat(save_format) if isinstance(save_format, str) else save_format
        self._path.mkdir(parents=True, exist_ok=True)

    def save(self, data: dict[str, list[Sample]]) -> dict[str, str]:
        """Save complete dataset, one file per task."""
        paths = {}
        for task, samples in data.items():
            paths[task] = self.save_task(task, samples)
        return paths

    def save_task(self, task: str, samples: list[Sample]) -> str:
        """Save samples for a single task."""
        path = self._task_path(task)
        self._write(path, samples)
        return str(path.resolve())

    def load_task(self, task: str) -> list[Sample]:
        """Load samples for a single task. Returns empty list if not found."""
        path = self._task_path(task)
        if not path.exists():
            return []
        return self._read(path)

    def _task_path(self, task: str) -> Path:
        """Get file path for a task."""
        safe_name = task.replace("/", "_").replace("\\", "_")
        return self._path / f"{safe_name}.{self._format.value}"

    def _write(self, path: Path, samples: list[Sample]) -> None:
        """Write samples to file."""
        if self._format == SaveFormat.JSON:
            with open(path, "w", encoding="utf-8") as f:
                json.dump(samples, f, indent=2, ensure_ascii=False, default=str)
        elif self._format == SaveFormat.JSONL:
            with open(path, "w", encoding="utf-8") as f:
                for s in samples:
                    f.write(json.dumps(s, ensure_ascii=False, default=str) + "\n")
        elif self._format == SaveFormat.CSV:
            with open(path, "w", encoding="utf-8", newline="") as f:
                writer = csv.DictWriter(f, fieldnames=["text", "label"])
                writer.writeheader()
                writer.writerows(samples)

    def load_all(self) -> dict[str, list[Sample]]:
        """Load all task files from the save directory."""
        data = {}
        ext = f".{self._format.value}"
        for path in self._path.glob(f"*{ext}"):
            task = path.stem
            data[task] = self._read(path)
        return data

    def _read(self, path: Path) -> list[Sample]:
        """Read samples from file."""
        if self._format == SaveFormat.JSON:
            with open(path, encoding="utf-8") as f:
                return json.load(f)
        elif self._format == SaveFormat.JSONL:
            samples = []
            with open(path, encoding="utf-8") as f:
                for line in f:
                    if line.strip():
                        samples.append(json.loads(line))
            return samples
        elif self._format == SaveFormat.CSV:
            with open(path, encoding="utf-8", newline="") as f:
                return list(csv.DictReader(f))



"""Loss functions for classification."""

import torch
import torch.nn as nn
import torch.nn.functional as F

from rapidfit.classifiers.config import LossConfig


class FocalLoss(nn.Module):
    """Focal loss for handling class imbalance."""

    def __init__(
        self,
        gamma: float = 2.0,
        weight: torch.Tensor | None = None,
        label_smoothing: float = 0.0,
    ) -> None:
        super().__init__()
        self.gamma = gamma
        self.weight = weight
        self.label_smoothing = label_smoothing

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        ce_loss = F.cross_entropy(
            inputs, targets,
            weight=self.weight,
            label_smoothing=self.label_smoothing,
            reduction="none",
        )
        pt = torch.exp(-ce_loss)
        focal_loss = ((1 - pt) ** self.gamma) * ce_loss
        return focal_loss.mean()


class TaskLoss:
    """Manages loss computation for multi-task learning."""

    def __init__(self, config: LossConfig) -> None:
        self.config = config
        self.class_weights: dict[str, torch.Tensor] = {}

    def set_class_weights(self, weights: dict[str, torch.Tensor]) -> None:
        self.class_weights = weights

    def compute(
        self,
        logits: torch.Tensor,
        labels: torch.Tensor,
        task: str,
    ) -> torch.Tensor:
        weight = self.class_weights.get(task)
        if weight is not None:
            weight = weight.to(logits.device)

        if self.config.use_focal_loss:
            loss_fn = FocalLoss(
                gamma=self.config.focal_gamma,
                weight=weight,
                label_smoothing=self.config.label_smoothing,
            )
        else:
            loss_fn = nn.CrossEntropyLoss(
                weight=weight,
                label_smoothing=self.config.label_smoothing,
            )

        loss = loss_fn(logits, labels)
        task_weight = self.config.task_weights.get(task, 1.0)
        return loss * task_weight

    def state_dict(self) -> dict[str, torch.Tensor]:
        return {k: v.cpu() for k, v in self.class_weights.items()}

    def load_state_dict(self, state: dict[str, torch.Tensor]) -> None:
        self.class_weights = state



"""LLM-based data augmenter."""

import random

import json_repair
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn

from rapidfit.augmenters.base import BaseAugmenter
from rapidfit.clients import ChatClient
from rapidfit.io import DataSaver
from rapidfit.prompts import load_prompt
from rapidfit.types import AugmentResult, ClassInstruction, Sample, SaveFormat, SeedData, TaskPrompts, WriteMode


class LLMAugmenter(BaseAugmenter):
    """Augmenter that uses LLM to generate synthetic samples."""

    def __init__(
        self,
        api_key: str,
        base_url: str | None = None,
        model_id: str = "gpt-4.1-mini",
        max_samples_per_task: int = 128,
        batch_size: int = 8,
        max_temperature: float = 0.9,
        save_path: str = "./saved",
        save_format: SaveFormat | str = SaveFormat.JSON,
        save_incremental: bool = True,
        write_mode: WriteMode | str = WriteMode.OVERWRITE,
    ) -> None:
        """
        Initialize LLM augmenter.

        Args:
            api_key: OpenAI API key.
            base_url: Optional custom API base URL.
            model_id: Model to use for generation.
            max_samples_per_task: Maximum samples per task after augmentation.
            batch_size: Number of samples to generate per batch.
            max_temperature: Maximum temperature for random sampling.
            save_path: Directory to save generated data.
            save_format: Output format (json, jsonl, csv).
            save_incremental: Save while generating instead of at the end.
            write_mode: How to handle existing files (overwrite or append).
        """
        self._client = ChatClient(api_key, base_url, model_id)
        self._max_samples = max_samples_per_task
        self._batch_size = batch_size
        self._max_temp = max_temperature
        self._analyze_template = load_prompt("analyze")
        self._generate_template = load_prompt("generate")
        self._console = Console()
        self._saver = DataSaver(save_path, save_format)
        self._save_incremental = save_incremental
        self._write_mode = WriteMode(write_mode) if isinstance(write_mode, str) else write_mode

    def augment(self, seed_data: SeedData) -> AugmentResult:
        """
        Augment dataset using LLM generation.

        Args:
            seed_data: Mapping of task names to sample lists.

        Returns:
            Mapping of task names to results with file paths and stats.
        """
        task_prompts = self._analyze_tasks(seed_data)
        return self._generate_samples(seed_data, task_prompts)

    def _analyze_tasks(self, seed_data: SeedData) -> TaskPrompts:
        """Analyze seed data and generate class prompts for each task."""
        task_prompts: TaskPrompts = {}

        with Progress(
            SpinnerColumn(),
            TextColumn("[bold blue]Analyzing[/] {task.description}"),
            TextColumn("[dim]calls:{task.fields[calls]} tokens:{task.fields[tokens]}[/]"),
            console=self._console,
        ) as progress:
            task = progress.add_task("", total=len(seed_data), calls=0, tokens=0)

            for task_name, samples in seed_data.items():
                progress.update(
                    task,
                    description=f"[cyan]{task_name}[/]",
                    calls=self._client.call_count,
                    tokens=self._client.total_tokens,
                )
                samples_text = "\n".join(
                    f"- [{s['label']}] {s['text']}" for s in samples
                )
                prompt = self._analyze_template.format(
                    task_name=task_name,
                    samples=samples_text,
                )
                response = self._client.complete(prompt)
                task_prompts[task_name] = json_repair.loads(response)
                progress.update(
                    task,
                    calls=self._client.call_count,
                    tokens=self._client.total_tokens,
                )
                progress.advance(task)

        return task_prompts

    def _generate_samples(
        self,
        seed_data: SeedData,
        task_prompts: TaskPrompts,
    ) -> AugmentResult:
        """Generate augmented samples for each task."""
        result: AugmentResult = {}
        augmented: SeedData = {}

        with Progress(
            SpinnerColumn(),
            TextColumn("{task.description}"),
            BarColumn(),
            TextColumn("[cyan]{task.completed}/{task.total}[/]"),
            TextColumn("[dim]calls:{task.fields[calls]} tokens:{task.fields[tokens]}[/]"),
            console=self._console,
        ) as progress:
            for task_name, samples in seed_data.items():
                existing = self._load_existing(task_name)
                existing_texts = {s["text"] for s in existing}
                base_samples = existing if existing else list(samples)

                class_instructions = task_prompts.get(task_name, {})
                labels = list(class_instructions.keys())

                if not labels:
                    augmented[task_name] = base_samples
                    path = self._saver.save_task(task_name, base_samples)
                    result[task_name] = {
                        "path": path,
                        "stats": self._compute_stats(base_samples),
                    }
                    continue

                counts = self._calculate_distribution(labels, len(base_samples))
                augmented[task_name] = base_samples
                total_samples = sum(counts.values())

                if total_samples <= 0:
                    path = self._saver.save_task(task_name, augmented[task_name])
                    result[task_name] = {
                        "path": path,
                        "stats": self._compute_stats(augmented[task_name]),
                    }
                    continue

                task_id = progress.add_task(
                    f"[bold]{task_name}[/]",
                    total=total_samples,
                    calls=self._client.call_count,
                    tokens=self._client.total_tokens,
                )

                for label in labels:
                    needed = counts[label]
                    if needed <= 0:
                        continue

                    progress.update(
                        task_id,
                        description=f"[bold]{task_name}[/] → [yellow]{label}[/]",
                        calls=self._client.call_count,
                        tokens=self._client.total_tokens,
                    )
                    new_samples = self._generate_for_class(
                        task_name, label, class_instructions[label], needed, existing_texts, progress, task_id
                    )
                    augmented[task_name].extend(new_samples)

                    if self._save_incremental:
                        self._saver.save_task(task_name, augmented[task_name])

                path = self._saver.save_task(task_name, augmented[task_name])
                result[task_name] = {
                    "path": path,
                    "stats": self._compute_stats(augmented[task_name]),
                }

        self._console.print(
            f"[bold green]Done![/] calls: {self._client.call_count}, tokens: {self._client.total_tokens}"
        )
        return result

    def _load_existing(self, task_name: str) -> list[Sample]:
        """Load existing samples if in append mode."""
        if self._write_mode == WriteMode.APPEND:
            return self._saver.load_task(task_name)
        return []

    def _compute_stats(self, samples: list[Sample]) -> dict:
        """Compute label distribution stats."""
        labels: dict[str, int] = {}
        for s in samples:
            labels[s["label"]] = labels.get(s["label"], 0) + 1
        return {"total": len(samples), "labels": labels}

    def _calculate_distribution(self, labels: list[str], existing_count: int = 0) -> dict[str, int]:
        """Calculate how many samples to generate per class for balance."""
        remaining = max(0, self._max_samples - existing_count)
        per_class = remaining // len(labels)
        return {label: per_class for label in labels}

    def _generate_for_class(
        self,
        task_name: str,
        label: str,
        instruction: ClassInstruction,
        count: int,
        existing_texts: set[str],
        progress: Progress,
        task_id: int,
    ) -> list[Sample]:
        """Generate samples for a specific class using batched iteration."""
        collected: set[str] = set()
        languages = instruction.get("languages", {})
        lang_str = ", ".join(f"{k} ({int(v*100)}%)" for k, v in languages.items())
        length = instruction.get("length", {})
        do_items = instruction.get("do", [])[:3]

        while len(collected) < count:
            batch_count = min(self._batch_size, count - len(collected))
            temp = random.uniform(0.5, self._max_temp)

            prompt = self._generate_template.format(
                task_name=task_name,
                label=label,
                languages=lang_str or "English",
                style=instruction.get("style", "")[:100],
                length_min=length.get("min", 10),
                length_max=length.get("max", 500),
                do_list="\n".join(f"- {d}" for d in do_items) if do_items else "- Be creative and natural",
                edge_cases="Vary tone, context, and vocabulary across samples.",
                count=batch_count,
            )

            response = self._client.complete(prompt, temperature=temp)
            texts = json_repair.loads(response)

            if isinstance(texts, list):
                for t in texts:
                    text = str(t).strip()
                    if text and text not in collected and text not in existing_texts:
                        collected.add(text)
                        existing_texts.add(text)
                        progress.update(
                            task_id,
                            calls=self._client.call_count,
                            tokens=self._client.total_tokens,
                        )
                        progress.advance(task_id)
                        if len(collected) >= count:
                            break

        return [{"text": t, "label": label} for t in collected]



"""Task-specific classification heads."""

import torch.nn as nn

from rapidfit.classifiers.config import ActivationType, HeadConfig

ACTIVATIONS: dict[ActivationType, type[nn.Module]] = {
    "gelu": nn.GELU,
    "relu": nn.ReLU,
    "silu": nn.SiLU,
    "tanh": nn.Tanh,
}


def build_head(
    input_size: int, num_labels: int, config: HeadConfig
) -> nn.Sequential:
    """Build a classification head with configurable architecture."""
    hidden_size = int(input_size * config.hidden_multiplier)
    activation_cls = ACTIVATIONS[config.activation]

    layers: list[nn.Module] = []
    in_features = input_size

    for _ in range(config.hidden_layers):
        layers.extend([
            nn.Linear(in_features, hidden_size),
            activation_cls(),
            nn.Dropout(config.dropout),
        ])
        in_features = hidden_size

    layers.append(nn.Linear(in_features, num_labels))
    return nn.Sequential(*layers)


class TaskHeads(nn.ModuleDict):
    """Container for task-specific classification heads."""

    def __init__(
        self,
        hidden_size: int,
        task_num_labels: dict[str, int],
        config: HeadConfig,
    ) -> None:
        heads = {
            task: build_head(hidden_size, num_labels, config)
            for task, num_labels in task_num_labels.items()
        }
        super().__init__(heads)
        self.config = config



"""Prompt template utilities."""

from pathlib import Path

_PROMPTS_DIR = Path(__file__).parent


def load_prompt(name: str) -> str:
    """
    Load a prompt template by name.

    Args:
        name: Prompt file name without extension.

    Returns:
        Prompt template content.
    """
    return (_PROMPTS_DIR / f"{name}.txt").read_text()



"""Pooling strategies for encoder outputs."""

import torch
import torch.nn as nn

from rapidfit.classifiers.config import PoolingStrategy


class Pooler(nn.Module):
    """Configurable pooling layer for transformer outputs."""

    def __init__(self, strategy: PoolingStrategy = "mean") -> None:
        super().__init__()
        self.strategy = strategy

    def forward(
        self, hidden_states: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        if self.strategy == "cls":
            return hidden_states[:, 0]
        if self.strategy == "max":
            return self._max_pool(hidden_states, attention_mask)
        return self._mean_pool(hidden_states, attention_mask)

    def _mean_pool(
        self, hidden_states: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        mask = attention_mask.unsqueeze(-1).expand(hidden_states.size()).float()
        summed = torch.sum(hidden_states * mask, dim=1)
        return summed / torch.clamp(mask.sum(dim=1), min=1e-9)

    def _max_pool(
        self, hidden_states: torch.Tensor, attention_mask: torch.Tensor
    ) -> torch.Tensor:
        mask = attention_mask.unsqueeze(-1).expand(hidden_states.size())
        hidden_states = hidden_states.masked_fill(~mask.bool(), float("-inf"))
        return hidden_states.max(dim=1).values



"""Configuration definitions for classifiers."""

from dataclasses import dataclass, field
from typing import Literal

PoolingStrategy = Literal["mean", "cls", "max"]
ActivationType = Literal["gelu", "relu", "silu", "tanh"]
TaskSampling = Literal["proportional", "equal", "sqrt"]
EarlyStoppingMetric = Literal["loss", "accuracy"]


@dataclass
class HeadConfig:
    """Task head architecture configuration."""

    hidden_layers: int = 1
    hidden_multiplier: float = 1.0
    activation: ActivationType = "gelu"
    dropout: float = 0.2


@dataclass
class EncoderConfig:
    """Encoder fine-tuning configuration."""

    freeze_epochs: int = 3
    unfreeze_layers: int | None = 4
    lr_multiplier: float = 0.1


@dataclass
class LossConfig:
    """Loss function configuration."""

    label_smoothing: float = 0.1
    use_class_weights: bool = True
    use_focal_loss: bool = False
    focal_gamma: float = 2.0
    task_weights: dict[str, float] = field(default_factory=dict)


@dataclass
class TrainingConfig:
    """Training hyperparameters."""

    model_name: str = "sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2"
    max_length: int = 128
    batch_size: int = 16
    epochs: int = 10
    learning_rate: float = 2e-5
    weight_decay: float = 0.01
    warmup_ratio: float = 0.1
    gradient_accumulation_steps: int = 1
    max_grad_norm: float = 1.0
    test_size: float = 0.1
    val_size: float = 0.1
    output_dir: str = "./training_output"
    save_path: str = "./model"


@dataclass
class EvalConfig:
    """Evaluation and early stopping configuration."""

    patience: int = 3
    metric: EarlyStoppingMetric = "accuracy"
    min_delta: float = 0.001


@dataclass
class MultiheadConfig:
    """Complete configuration for MultiheadClassifier."""

    training: TrainingConfig = field(default_factory=TrainingConfig)
    head: HeadConfig = field(default_factory=HeadConfig)
    encoder: EncoderConfig = field(default_factory=EncoderConfig)
    loss: LossConfig = field(default_factory=LossConfig)
    eval: EvalConfig = field(default_factory=EvalConfig)
    pooling: PoolingStrategy = "mean"
    task_sampling: TaskSampling = "proportional"

    @classmethod
    def from_dict(cls, data: dict) -> "MultiheadConfig":
        """Create config from dictionary."""
        return cls(
            training=TrainingConfig(**data.get("training", {})),
            head=HeadConfig(**data.get("head", {})),
            encoder=EncoderConfig(**data.get("encoder", {})),
            loss=LossConfig(**data.get("loss", {})),
            eval=EvalConfig(**data.get("eval", {})),
            pooling=data.get("pooling", "mean"),
            task_sampling=data.get("task_sampling", "proportional"),
        )

    def to_dict(self) -> dict:
        """Convert config to dictionary."""
        from dataclasses import asdict
        return asdict(self)


DEFAULT_CONFIG = MultiheadConfig()



"""Chat completion client."""

from openai import OpenAI


class ChatClient:
    """Chat completion client using OpenAI-compatible API."""

    def __init__(
        self,
        api_key: str,
        base_url: str | None = None,
        model_id: str = "gpt-4.1-mini",
    ) -> None:
        """
        Initialize OpenAI client.

        Args:
            api_key: OpenAI API key.
            base_url: Optional custom base URL for API.
            model_id: Model identifier to use.
        """
        self.model_id = model_id
        self._client = OpenAI(api_key=api_key, base_url=base_url)
        self.call_count = 0
        self.total_tokens = 0

    def complete(self, prompt: str, system: str = "", temperature: float = 0.7) -> str:
        """
        Get completion from the model.

        Args:
            prompt: User prompt.
            system: Optional system message.
            temperature: Sampling temperature.

        Returns:
            Model response text.
        """
        messages = []
        if system:
            messages.append({"role": "system", "content": system})
        messages.append({"role": "user", "content": prompt})

        response = self._client.chat.completions.create(
            model=self.model_id,
            messages=messages,
            temperature=temperature,
        )
        self.call_count += 1
        if response.usage:
            self.total_tokens += response.usage.total_tokens
        return response.choices[0].message.content or ""



"""Base classifier interface."""

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from rapidfit.types import AugmentResult, ClassifierConfig, Prediction, SeedData


class BaseClassifier(ABC):
    """Abstract base class for classifiers."""

    def __init__(self, config: ClassifierConfig | None = None) -> None:
        self._config = config or {}
        self._is_trained = False

    @property
    def is_trained(self) -> bool:
        return self._is_trained

    @abstractmethod
    def train(
        self,
        data: SeedData | AugmentResult | None = None,
        data_save_dir: str | None = None,
    ) -> None:
        """
        Train classifier on data.

        Args:
            data: Either SeedData dict or AugmentResult from augmenter.
            data_save_dir: Path to directory containing saved augmented data.
        """
        pass

    @abstractmethod
    def predict(self, texts: list[str], task: str) -> list[Prediction]:
        """
        Predict labels for texts on a specific task.

        Args:
            texts: List of input texts.
            task: Task name to use for prediction.

        Returns:
            List of predictions with labels and confidence.
        """
        pass

    @abstractmethod
    def save(self, path: str | Path) -> None:
        """Save trained model to disk."""
        pass

    @abstractmethod
    def load(self, path: str | Path) -> None:
        """Load trained model from disk."""
        pass

    def _resolve_data(
        self,
        data: SeedData | AugmentResult | None,
        data_save_dir: str | None,
    ) -> SeedData:
        """Resolve input to SeedData format."""
        if data_save_dir:
            return self._load_from_dir(data_save_dir)

        if not data:
            return {}

        first_value = next(iter(data.values()), None)
        if isinstance(first_value, list):
            return data

        return self._load_from_augment_result(data)

    def _load_from_dir(self, dir_path: str) -> SeedData:
        """Load all task data from a directory."""
        from rapidfit.io import DataSaver
        from rapidfit.types import SaveFormat

        path = Path(dir_path)
        if not path.exists():
            raise ValueError(f"Directory not found: {dir_path}")

        for fmt in SaveFormat:
            files = list(path.glob(f"*.{fmt.value}"))
            if files:
                return DataSaver(dir_path, fmt).load_all()
        return {}

    def _load_from_augment_result(self, data: AugmentResult) -> SeedData:
        """Load data from AugmentResult paths."""
        import csv
        import json

        resolved: SeedData = {}
        for task, result in data.items():
            file_path = Path(result.get("path", ""))
            if not file_path.exists():
                continue
            if file_path.suffix == ".jsonl":
                with open(file_path, encoding="utf-8") as f:
                    resolved[task] = [json.loads(line) for line in f if line.strip()]
            elif file_path.suffix == ".json":
                with open(file_path, encoding="utf-8") as f:
                    resolved[task] = json.load(f)
            elif file_path.suffix == ".csv":
                with open(file_path, encoding="utf-8") as f:
                    resolved[task] = list(csv.DictReader(f))
        return resolved
